/****************************************************************************/
/*      Motorola Time Processor Unit microcode assembler                    */
/****************************************************************************/
/*                                                                          */
/*      Copyright 1998, (c) by Josef Wolf <jw@raven.inka.de>                */
/*                                                                          */
/*      This file is under the GNU General Public License.                  */
/*      See LICENSE for details.                                            */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*      This creates the C interface header file.                           */
/*                                                                          */
/****************************************************************************/

# include "tpuasm.h"

/* The TPU can handle at most...
 */
# define MAXFUNCS 16	/* ... 16 functions, */
# define HSRNUM 3	/* ... 3 host service requests per function */
# define PARAMS 8	/* ... 8 parameters per function */

/* prepare for a better handling of parameter comments -- not used yet
 */
typedef struct comment_s {
  char *text;
  struct comment_s *next;
} comment_t;

typedef struct mac_s {
  char *name;
  char *val;
  char *comment;
  struct mac_s *next;
} mac_t;

typedef struct func_s {
  char *name;
  char *hsrname[HSRNUM];
  char *params[PARAMS];
  char *comment[PARAMS];
  mac_t *macros;
} func_t;

static int max_pc = 0;		/* No memory is used yet ... */
static int min_ent = 15;	/* No etries defined yet ... */
static int mem_end = 255;	/* let the 256-word-TPU be the default */
static int the_func;		/* the function currently processed */
static char *my_name;		/* name of the assembler */
static func_t funcs[MAXFUNCS];	/* storage for the collected information */

void cif_memend (int end)
{
  mem_end = end-1;	/* We'll need this one later */
}

void cif_touch_ent (int ent)
{
  if (ent < min_ent) min_ent = ent;	/* remember least used entry... */
}
void cif_touch_pc (int pc)
{
  if (pc > max_pc) max_pc = pc;		/* ... and highest used address */
}

void cif_func (char *name, int funcnr)
{
  if (funcnr<0 || funcnr>=MAXFUNCS || !name) return;	/* sanity */
  funcs[funcnr].name = name;	/* remember name of function */
  the_func = funcnr;		/* and let it be the current one. */
}

void cif_entry (int funcnr, char *name, Bool hsr0, Bool hsr1)
{
  char *p;
  int i=0, j, k;

  if (funcnr<0 || funcnr>=MAXFUNCS) return;	/* sanity */
  the_func = funcnr;			/* make it current */
  if (!name) return;			/* more sanity */
  if (hsr0==True) i|=1;			/* calc the num of the HSR */
  if (hsr1==True) i|=2;
  if (!i) return;			/* even more sanity */
  for (j=0; j<HSRNUM; j++) {
    for (k=0; k<MAXFUNCS; k++) {
      if ((p=funcs[k].hsrname[j])) {
        if (!strcmp (p, name)) {
	  return;			/* Already defined */
	}
      }
    }
  }
  funcs[funcnr].hsrname[i-1] = name;	/* remember name of the HSR */
}

void cif_macro (char *name, char *val, char *comment)
{
  int num;
  mac_t *mac, *tmp_mac, *tmp_mac1;

  if (!val || !name) return;		/* sanity */
  if (isdigit(val[3]) && (!val[4]) && !strncmp (val, "prm", 3)) {
    num = val[3] - '0';			/* this is a parameter definition */
    if (num>=0 && num<PARAMS) {		/* remember if valid */
      funcs[the_func].params[num] = name;
      funcs[the_func].comment[num] = comment;
    }
  } else {
    /* ordinary macro... */
    if (!(mac = malloc (sizeof(*mac)))) FatalError ((E_OUTMEM));
    mac->name = name;
    mac->val = val;
    mac->comment = comment;
    mac->next = NULL;
    /* ... Just remember it */
    if ((tmp_mac1 = funcs[the_func].macros)) {
      for (tmp_mac=tmp_mac1; tmp_mac->next; tmp_mac=tmp_mac->next)
	;
      tmp_mac->next = mac;
    } else {
      funcs[the_func].macros = mac;
    }
  }
}

/* calculate memory location of specified entry
 */
static int ent_off (int func, int ent_nr)
{
 return (mem_end - 127 + (func*16 + ent_nr) / 2);
}

/* format and write comment of a macro definition
 */
static void put_comment (FILE *fp, int firstlen, int secondlen, char *comment)
{
  char *p, sav, buf[200], *t="\t";

  /* create indentation string for the comment. This indentation is used
   * for every line except the first one, because the first one already
   * contains the defintion.
   */
  p="                                                                      ";
  strncpy (buf, p, firstlen); buf[firstlen] = '\0';
  strcat (buf, "\t");
  if (secondlen) {
    strncat (buf, p, secondlen); buf[firstlen+secondlen+1] = '\0';
    strcat (buf, "\t");
  }

  if (comment && *comment) {
    while ((p = strchr (comment, '\n'))) {	/* line by line... */
      while (isspace (*comment)) comment++;	/* ... skip original space */
      sav = *p; *p = '\0';			/* ... terminate line */
      fprintf (fp, "%s/* %s */\n", t, comment);	/* ... indent and print */
      t=buf;				/* ... prepare indentation... */
      comment = p+1; *p = sav;		/* ... and contents for next line */
    }
    /* Don't forget last line
     */
    while (isspace (*comment)) comment++;
    if (*comment) fprintf (fp, "%s/* %s */", t, comment);
  }
  fprintf (fp, "\n");
}

void cif_flush (char *srcnam, FILE *fp)
{
  char *name;
  mac_t *macro;
  long chsum = 0;
  int i, j, k, v, flag;

  if (!strcmp (srcnam, "-")) srcnam = "stdin";

  /* some prosa ...
   */
  fprintf (fp, "/* ATTENTION! DON'T CHANGE THIS FILE BY HAND\n");
  fprintf (fp, " *\n");
  fprintf (fp, " * This file was generated automatically from '%s'\n", srcnam);
  fprintf (fp, " * by %s\n", my_name);
  fprintf (fp, " * If you _really_ need to make some changes, then '%s'\n",
	   srcnam);
  fprintf (fp, " * is the file you are looking for.\n");
  fprintf (fp, " */\n\n");
  fprintf (fp, "# ifndef TPU_MICROCODE_IFACE\n");
  fprintf (fp, "# define TPU_MICROCODE_IFACE\n\n");

  /* Define all the functions and the HSR's/non-parameter-macros for those
   * functions. This is straight forward...
   */
  for (i=0; i<MAXFUNCS; i++) {
    if ((name = funcs[i].name)) {
      fprintf (fp, "/*\n * Define the %s function.\n */\n", name);
      fprintf (fp, "#define TPU_%s\t%d\n", name, i);
      flag = 1;
      for (j=0; j<HSRNUM; j++) {
	if (funcs[i].hsrname[j]) {
	  if (flag) fprintf (fp, "/*\n * Host services for %s\n */\n", name);
	  flag = 0;
	  fprintf (fp, "#define TPU_HSR_%s\t%d\n", funcs[i].hsrname[j], j+1);
	}
      }
      flag = 1;
      for (macro = funcs[i].macros; macro; macro=macro->next) {
	if (flag) fprintf (fp, "/*\n * Definitions for %s\n */\n", name);
	flag = 0;
	fprintf (fp,"#define TPUD_%s_%s\t%s", name, macro->name, macro->val);
	put_comment (fp, 14+strlen(macro->name), strlen(macro->val),
		     macro->comment);
      }
      fprintf (fp, "\n\n");
    }
  }

  /* define the structure for the parameter entries
   */
  fprintf (fp, "/* give direct access to tpu registers. access them via:\n");
  fprintf (fp, " *              tpureg[channel].funct.PAR.w    = someword;\n");
  fprintf (fp, " *  otherword = tpureg[channel].funct.PAR.w;\n");
  fprintf (fp, " *              tpureg[channel].funct.PAR.b.hi = somebyte;\n");
  fprintf (fp, " *  otherbyte = tpureg[channel].funct.PAR.b.low;\n");
  fprintf (fp, " * or some such\n");
  fprintf (fp, " */\n");
  fprintf (fp, "typedef union wordreg_u {\n");
  fprintf (fp, " volatile unsigned short w;\n");
  fprintf (fp, " struct b_s {\n");
  fprintf (fp, "  volatile unsigned char h;\n");
  fprintf (fp, "  volatile unsigned char l;\n");
  fprintf (fp, " } b;\n");
  fprintf (fp, "} wordreg_t;\n\n");

  /* define the structure which contains all parameter entries of all
   * known functions
   */
  fprintf (fp, "typedef union tpureg_u {\n");
  for (i=0; i<MAXFUNCS; i++) {
    if ((name = funcs[i].name)) {
      fprintf (fp, " struct TPU_%s_s {\n", name);
      for (j=0; j<PARAMS; j++) {
	if (funcs[i].params[j]) {
	  fprintf (fp, "  wordreg_t %s;", funcs[i].params[j]);
	  put_comment (fp, 13+strlen(funcs[i].params[j]), 0,
		       funcs[i].comment[j]);
	} else {
	  fprintf (fp, "  wordreg_t _D_O_N_T___U_S_E___%d;\n", j);
	}
      }
      fprintf (fp, " } %s;\n", name);
    }
  }
  fprintf (fp, "} tpureg_t;\n\n");
  fprintf (fp, "# endif\n\n");

  /* some helper definitions for the TPU initialization code
   */
  fprintf (fp, "# ifdef TPUGLUE\n/*\n * The TPU code\n */\n");
  fprintf (fp, "# define TPULOWEND      %d\n", max_pc);
  fprintf (fp, "# define TPUHIEND       %d\n", mem_end);
  fprintf (fp, "# define TPUHISTART     %d\n", ent_off (min_ent, 0));

  /* The memory contents are split into two parts to save memory when
   * large parts are not used
   */

  /* the microcode...
   */
  fprintf (fp, "static unsigned char tpuprog1[] = {\n");
  for (i=0; i<=max_pc; i+=2) {
    fprintf (fp, " /* 0x%04x */\t", i);
    for (j=0; j<2; j++) {
      for (k=0; k<4; k++) {
	v = (int) (((tpuram[i+j].value) >>((3-k)*8)) & 0xff);
	chsum += v;
	fprintf (fp, "0x%02x", v);
	if ((k<3) || (!j) || ((i+j)<max_pc)) fprintf (fp, ", ");
      }
      if (!j) fprintf (fp, " ");
    }
    fprintf (fp, "\n");
  }
  fprintf (fp, "};\n");

  /* the entry points...
   */
  fprintf (fp, "static unsigned char tpuprog2[] = {\n");
  for (i=min_ent; i<MAXFUNCS; i++) {
    for (j=0; j<16; j++) {
      v = entries[i*16+j].value;
      chsum += (v & 0xff) + (v >> 8);
      if (!(j%4)) fprintf (fp, " /* 0x%04x */\t", ent_off (i, j));
      fprintf (fp, "0x%02x, 0x%02x", (int) ((v>>8)&0x0ff), (int)(v&0x0ff));
      if (i != (MAXFUNCS-1) || j<15) fprintf (fp, ", ");
      if ((j%4)==1) fprintf (fp, " ");
      if ((j%4)==3) fprintf (fp, "\n");
    }
  }
  fprintf (fp, "};\n");
  fprintf (fp, "#define TPUCHSUM 0x%06lx\n", chsum);
  fprintf (fp, "#endif\n");
  fprintf (fp, "/* EOF */\n");
}

void cif_init (char *progname)
{
  int i, j;

  my_name = progname;

  for (i=0; i<MAXFUNCS; i++) {
    funcs[i].name = NULL;
    funcs[i].macros = NULL;
    for (j=0; j<HSRNUM; j++) {
      funcs[i].hsrname[j] = NULL;
    }
    for (j=0; j<PARAMS; j++) {
      funcs[i].params[j] = NULL;
    }
  }
}
