/****************************************************************************/
/*			  Motorola Time Processor Unit microcode assembler 				*/
/****************************************************************************/
/*																			*/
/*	Copyright 1998, (c) Bendor Research Pty. Ltd.							*/
/*																			*/
/*	This file is under the GNU General Public License.						*/
/*	See LICENSE for details.												*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	This file contains the directive processing of the assembler.			*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	Changelog:																*/
/*																			*/
/*	30/04/98	Zoltan Kocsi	Alpha version								*/
/*																			*/
/****************************************************************************/

#include "tpuasm.h"

/****************************************************************************/
/*							Global data										*/
/****************************************************************************/

int		maxmem = 0;						/* Defines the max. TPU memory		*/
int		def_function;					/* Default function					*/
int		def_func_set = False;			/* Default function is valid		*/
ENTRY	entry;							/* Global entry descriptor			*/

/****************************************************************************/
/*					Local definitions and type								*/
/****************************************************************************/

typedef struct {

	Bool	hsr1, 
			hsr0, 
			lsr, 
			mtsr, 
			pin, 
			flag0;
	
} TEMPL;

typedef	struct	s_incf {

	struct	s_incf	*next;
	char			*name;

} INCF;

/****************************************************************************/
/*						Local function prototypes							*/
/****************************************************************************/

static	Bool	AlreadyIncluded( char *name );
static	void	TestValBool( ENTRY *entry, Bool n );

/****************************************************************************/
/*							Local data										*/
/****************************************************************************/

static	MACRO	*macros = NULL;
static	INCF	*incl   = NULL;
static	INCF	*path	= NULL;
static	int		maxilen = 0;

static	const TEMPL template[] = {

	{ False, True,  	DontCare, DontCare, False,    DontCare },
	{ False, True,  	DontCare, DontCare, True,     DontCare },
	{ True,  False, 	DontCare, DontCare, DontCare, DontCare },
	{ True,  True,  	DontCare, DontCare, DontCare, DontCare },
	{ False, False, 	False,    True,     False,    False    },
	{ False, False, 	False,    True,     False,    True     },
	{ False, False, 	False,    True,     True,     False    },
	{ False, False, 	False,    True,     True,     True     },
	{ False, False, 	True,     False,    False,    False    },
	{ False, False, 	True,     False,    False,    True     },
	{ False, False, 	True,     False,    True,     False    },
	{ False, False, 	True,     False,    True,     True     },
	{ False, False, 	True,     True,     False,    False    },
	{ False, False, 	True,     True,     False,    True     },
	{ False, False, 	True,     True,     True,     False    },
	{ False, False, 	True,     True,     True,     True     },
};

/****************************************************************************/
/*							Public functions								*/
/****************************************************************************/

/*-----------------------------------------------------------------------
	Directive: %macro
-------------------------------------------------------------------------*/
/*
*	Macro definition
*/

void	MacroDef( char *name, char *value )
{
MACRO	*p;

	if ( ( p = MacroSearch( name ) ) ) {
	
 		if (strcmp (p->value, value)) {
	
			IssueError(( E_MULMAC, name ));
		}
		else
			return;
	}
		
	if ( ! ( p = malloc( sizeof( MACRO ) ) ) ) {
		 
		FatalError(( E_OUTMEM ));
	}
	
	if ( ! ( p->name = malloc( strlen( name ) + 1 ) ) ) {
		 
		FatalError(( E_OUTMEM ));
	}
	
	if ( ! ( p->value = malloc( strlen( value ) + 1 ) ) ) {
		 
		FatalError(( E_OUTMEM ));
	}

	strcpy( p->name,  name  );
	strcpy( p->value, value );
	p->next  = macros;
	macros   = p;
}

/*
*	Looks up a macro by name
*/

MACRO	*MacroSearch( char *name )
{
MACRO	*p;

	for ( p = macros ; p ; p = p->next )
	
		if ( ! strcmp( p->name, name ) )
		
			break;
	
	return( p );
}

void	DoOrg( int val )
{
	if ( val < 0 || val > maxmem ) {
	
		IssueError(( E_ORGVAL, val, maxmem ));
	}
	else {
	
		if ( pc > val ) IssueWarning(( E_ORGBCK, val, pc ));
		pc = val;
	}
}

/*-----------------------------------------------------------------------
	Directive: %type
-------------------------------------------------------------------------*/

/*
*	Sets the type of the TPU we work with
*/

void	DoSetType( int n )
{
	if ( maxmem && maxmem != n ) {
	
		IssueError(( E_MEMTYP, maxmem ));
	}
	else {
		maxmem = n;
		cif_memend (n);
	}
}

/*-----------------------------------------------------------------------
	Directive: %type
-------------------------------------------------------------------------*/

/*
*	Inserts a page break into the listfile
*/

void	DoPage( void )
{
	/* It doesn't :-) */
}

/*-----------------------------------------------------------------------
	Directive: %entry
-------------------------------------------------------------------------*/

/*
*	Initialises the %entry directive assembler's data
*/

void	EntryInit( void )
{
	entry.error = False;
	entry.fset  = False;
	entry.name  = NULL;
	entry.fixed = Unknown;
	entry.mena  = Unknown;
	entry.hsr0  = Unknown;
	entry.hsr1  = Unknown;
	entry.lsr   = Unknown;
	entry.mtsr  = Unknown;
	entry.pin   = Unknown;
	entry.flag0 = Unknown;
	entry.reg_p = Unknown;
	entry.magic = False;
}

/*
*	Finalizes an entry
*/

void	DoEntry( void )
{
int		i;
int		base;

	if ( ! entry.fset ) {
	
		if ( def_func_set ) {
		
			entry.fset = True;
			entry.func = def_function;
		}
		else {
		
			IssueError(( E_ENFUNC ));
			entry.error = True;
		}
	}
	
	if ( entry.fixed == Unknown ) {
	
		IssueError(( E_ENADDR ));
		entry.error = True;
	}
	
	if ( entry.mena == Unknown ) {
	
		entry.mena = True;	/* Default: enable match. */
	}
	
	/* If no preload was specified, the default is to load
	   parameter 7 to the DIOB (all bits are 1) */
		
	if ( entry.reg_p == Unknown ) {
	
		entry.reg_p = False;
		entry.param = 7;
		entry.magic = True;
	}
	
	/* If we had an error, then we can't calculate the address so we'd
	   better return. */
	
	if ( entry.error ) {
	
		EntryInit();
		return;
	}
	
	if ( ! maxmem ) {
	
		IssueError(( E_MEMUNS ));
		EntryInit();
		return;
	}
	
	/* Calculate all the entry points that match current config */
		
	base = entry.func * 16;
	
	if ( entry.hsr0  == Unknown ) entry.hsr0  = DontCare;
	if ( entry.hsr1  == Unknown ) entry.hsr1  = DontCare;
	if ( entry.lsr   == Unknown ) entry.lsr   = DontCare;
	if ( entry.mtsr  == Unknown ) entry.mtsr  = DontCare;
	if ( entry.pin   == Unknown ) entry.pin   = DontCare;
	if ( entry.flag0 == Unknown ) entry.flag0 = DontCare;
	
	/* Check for invalid combinations */
	
	if ( entry.hsr0 == False && entry.hsr1 == False &&
		 entry.lsr  == False && entry.mtsr == False ) {
		 
		IssueError(( E_INVENT ));
		EntryInit();
		return;
	}
			
	for ( i = 0 ; i < 16 ; i++ ) {
	
		if ( entry.hsr0 != DontCare && entry.hsr0 != template[ i ].hsr0 ) 
		
			continue;
			
		if ( entry.hsr1 != DontCare && entry.hsr1 != template[ i ].hsr1 ) 
		
			continue;
		
		if ( template[ i ].lsr == DontCare ) {
		
			if ( entry.lsr != DontCare ) {
			
				IssueWarning(( E_LSRIGN ));
			}
		}
		else if ( entry.lsr != DontCare && entry.lsr != template[ i ].lsr )
		
			continue;
		
		if ( template[ i ].lsr == DontCare ) {
		
			if ( entry.lsr != DontCare ) {
			
				IssueWarning(( E_LSRIGN ));
			}
		}
		else if ( entry.lsr != DontCare && entry.lsr != template[ i ].lsr )
		
			continue;
		
		if ( template[ i ].flag0 == DontCare ) {
		
			if ( entry.flag0 != DontCare ) {
			
				IssueWarning(( E_FLGIGN ));
			}
		}
		else if ( entry.flag0 != DontCare && entry.flag0 != template[i].flag0 )
		
			continue;
		
		if ( template[ i ].pin == DontCare ) {
		
			if ( entry.pin != DontCare ) {
			
				IssueWarning(( E_PINIGN ));
			}
		}
		else if ( entry.pin != DontCare && entry.pin != template[ i ].pin )
		
			continue;
		
		if ( template[ i ].mtsr == DontCare ) {
		
			if ( entry.mtsr != DontCare ) {
			
				IssueWarning(( E_MTSIGN ));
			}
		}
		else if ( entry.mtsr != DontCare && entry.mtsr != template[ i ].mtsr )
		
			continue;
		
		if ( entries[ base+i ].valid != Unknown ) {
		
			IssueError(( E_MEMFUL, base + i, E_MEMTEN ));
			entries[ base+i ].valid = False;			
			continue;
		}
		
		entry.srcn  = sourcename;
		entry.srcl  = src_lineno;
		entry.valid = True;
		entries[ base+i ] = entry;
	}

	cif_touch_ent (entry.func);
	cif_entry (entry.func, entry.name, entry.hsr0, entry.hsr1 );
	  
	EntryInit();
}

/*
*	Sets the function number for an entry
*/

void	EntryFunc( int n )
{
	if ( n < 0 || n > 15 ) {
	
		entry.error = True;
		IssueError(( E_EFUNNO ));
	}
	else if ( entry.fset != False ) {
	
		entry.error = True;
		IssueError(( E_EFUNST ));
	}
	else {
	
		entry.fset = True;
		entry.func = n;
	}
}

/*
*	Sets the name for an entry
*/

void	EntryName( IDENT *p )
{
	entry.name = p->name;
}

/*
*	Sets the start address for an entry
*/

void	EntryStart( Bool constant, int value )
{
	if ( entry.fixed != Unknown ) {
	
		IssueError(( E_ESTRTM ));
		entry.error = True;
		return;
	}
	
	if ( constant ) {
	
		if ( value < 0 || value > maxmem ) {
		
			IssueError(( E_EINVST, value, maxmem ));
			entry.error = True;
			return;
		}
		
		entry.fixed = True;
		entry.start = value;
	}
	else {
	
		entry.fixed = False;
		entry.start = value;
	}
}

/*
*	Sets the match enable control for an entry
*/

void	EntryMatch( Bool n )
{
	if ( entry.mena != Unknown )
	
		IssueWarning(( E_EMENAM ));
	
	entry.mena = n;
}

/*
*	Sets the register preload register for an entry
*/

void	EntryRegp( Bool preg )
{
	if ( entry.reg_p != Unknown )
	
		IssueWarning(( E_ERAMSM ));
	
	entry.reg_p = preg;
}

/*
*	Sets the register preload memory address for an entry
*/

void	EntryPar( int n )
{
	entry.param = n;
}

/*
*	Sets an entry condition field for an entry
*/

void	EntryCond( Ematch which, Bool val )
{
Bool	*field;
static	const	char	*names[] = { MATCHNAME };

	switch ( which ) {
	
		case EmatchHsr0:
		
			field = &entry.hsr0;
			TestValBool( &entry, val );
			break;
		
		case EmatchHsr1:
		
			field = &entry.hsr1;
			TestValBool( &entry, val );
			break;
			
		case EmatchLsr:
		
			field = &entry.lsr;
			TestValBool( &entry, val );
			break;

		case EmatchMtsr:
		
			field = &entry.mtsr;
			TestValBool( &entry, val );
			break;
		
		case EmatchPin:
		
			field = &entry.pin;
			TestValBool( &entry, val );
			break;
		
		case EmatchFlag0:
		
			field = &entry.flag0;
			TestValBool( &entry, val );
			break;
		
		case EmatchLsrX:
		
			field = &entry.lsr;
			break;

		case EmatchMtsrX:
		
			field = &entry.mtsr;
			break;
		
		case EmatchPinX:
		
			field = &entry.pin;
			break;
		
		case EmatchFlag0X:
		
			field = &entry.flag0;
			break;
		
		default:
		
			FatalError(( E_INTERN ));
			exit (EXIT_FAILURE); /* avoid gcc warning */
	}
	
	if ( *field != Unknown )
		
		IssueWarning(( E_EFLDST, names[ which ] ));
	
	*field = val;
}

void	FixupEntry( void )
{
IDENT	*id;
int		i;

	for ( i = 0 ; i < MAXENT ; i++ ) {
	
		if ( entries[ i ].valid == Unknown ) continue;
		if ( entries[ i ].error == True    ) continue;
		
		/* If the entry uses a symbolic address, fix it now */
		
		if ( entries[ i ].fixed == False ) {
		
			if ( ! ( id = IdentById( entries[ i ].start ) ) ) {
			
				FatalError(( E_INTERN ));
			}
			
			if ( id->is_set == False ) {
			
				FixupError( entries[ i ].srcn, entries[ i ].srcl,
							( E_UNDSYM, id->name ));
				entries[ i ].error = True;
				continue;
			}
			
			entries[ i ].start = id->value;
		}
		
		/* Calculate the entry. The format is:
		   15-13 : preload parameter source number
		      12 : Match enable (0:enabled)
			  11 : preload target (0:P 1:DIOB)
		   10- 9 : 00 (two 0 bits) by the manual; magic in reality ...
		    8- 0 : Entry start address (in longwords)
		*/
		
		entries[ i ].value = ( entries[ i ].param << 13 ) |
							 ( entries[ i ].magic ? 0x0600 : 0 ) |
							 ( entries[ i ].mena  ? 0x1000 : 0 ) |
							 ( entries[ i ].reg_p ? 0 : 0x0800 ) |
							 entries[ i ].start;
		
		/* Check if the entry's memory address is occupied by 
		   an instruction */
		
		if ( tpuram[ maxmem - 128 + i / 2 ].valid != Unknown ) {
		
			FixupError( entries[ i ].srcn, entries[ i ].srcl,
						( E_MEMFUL, maxmem - 128 + i / 2, E_MEMTIN ));
			entries[ i ].error = True;
		}
	}
}
				
/*-----------------------------------------------------------------------
	Directive: %include
-------------------------------------------------------------------------*/

/*
*	Sets the default function number for an included file
*/

void SetDefaultFunc( int n )
{
	if ( n < 0 || n > 15 ) {
	
		IssueError(( E_EFUNNO ));
	}
	else {
	
		def_function = n;
		def_func_set = True;
	}
}

/*
*	Includes a file
*/

void DoInclude( char *name )
{
INCF	*p;
char	*file;
static	const	char	eol[] = "\n";
INCF	*in;
FILE	*f;

	if ( ! ( file = malloc( maxilen + strlen( name ) + 2 ) ) )
	
		FatalError(( E_OUTMEM ));
	
	for ( in = path ; in ; in = in->next ) {
	
		strcpy( file, in->name );
		strcat( file, "/" );
		strcat( file, name );
		if ( ( f = fopen( file, "r" ) ) ) break;
	}
	
	if ( ! in ) {
	
		IssueError(( E_FOPENE, name ));
		free( file );
		return;
	}
	
	if ( AlreadyIncluded( file ) ) {
	
		IssueWarning(( E_MULINC, file ));
		free( file );
		return;
	}
	
	if ( ! ( p = malloc( sizeof( INCF ) ) ) ) {
		 
		FatalError(( E_OUTMEM ));
	}
	else {
	
		p->name = file;
		p->next = incl;
		incl = p;
	}
	
	Token( eol, False );
	sourcename = file;
	SrcOpen( True, file, f );
}

/*
*	Adds an include file path to the existing paths
*/

void	AddIncludePath( char *s )
{
INCF	*d;
INCF	*sc;
int		n;

	if ( ! ( d = malloc( sizeof( INCF ) ) ) ) 
	
		FatalError(( E_OUTMEM ));
	
	if ( ( n = strlen( s ) ) > maxilen ) maxilen = n;
	
	d->next = NULL;
	d->name = s;
	
	if ( ! 	path ) {
	
		path = d;
	}
	else {
	
		for ( sc = path ; sc->next ; sc = sc->next );
		sc->next = d;
	}
}

/****************************************************************************/
/*							Local  functions								*/
/****************************************************************************/

/*
*	Checks if an include file has already been included
*/

static	Bool	AlreadyIncluded( char *name )
{
INCF	*p;

	for ( p = incl ; p ; p = p->next ) 
	
		if ( ! strcmp( p->name, name ) )
		
			return( True );
	
	return( False );
}

/*
*	Tests if a boolean is indeed true or false
*/

static	void	TestValBool( ENTRY *entry, Bool n )
{
	if ( n != False && n != True ) {
		
		entry->error = True;
		IssueError(( E_CSTBOL ));
	}
}
	
