/****************************************************************************/
/*			  Motorola Time Processor Unit microcode assembler 				*/
/****************************************************************************/
/*																			*/
/*	Copyright 1998, (c) Bendor Research Pty. Ltd.							*/
/*																			*/
/*	This file is under the GNU General Public License.						*/
/*	See LICENSE for details.												*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	This file contains the definitions of the microintruction formats		*/
/*	and the subinstruction field encoding									*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	Changelog:																*/
/*																			*/
/*	30/04/98	Zoltan Kocsi	Alpha version								*/
/*																			*/
/****************************************************************************/

#ifndef	__INSN_H__
#define	__INSN_H__

/****************************************************************************/
/*			Simple instruction subfield setting functions					*/
/****************************************************************************/

/*

The TPU has 5 instruction formats which has the following structure:

Format 1
--------

31:30	00
   29	RW			RAM			Read Write
28:25	T1ABS		AU			A Bus Source ( in T1 )
24:21	T3ABD		AU			A Bus Destination ( in T3 )
20:19	SHF			AU			SHiFter control
   18	SRC			AU			Shift Register Control
   17	CCL			AU			Condition Code Latch vontrol
16:14	T1BBS		AU			B Bus Source ( in T1 )
   13	CIN			AU			B bus Carry IN control
   12	BINV		AU			B bus INVert control
11: 9	IOM			RAM			Input/Output Mode control
 8: 2	AID			RAM			Address Immediate
 1: 0	DECEND		SEQ			DECrementor/END control
 
Format 2
--------

31:30	00
   29	ERW			CHAN		Event Register Write control
28:25	T1ABS		as above
24:21	T3ABD		as above
20:19	SHF			as above
   18	TDL			CHAN		Transition Detect Latch negation control
   17	MRL			CHAN		Match Recognition Latch negation control
16:14	T1BBS		as above
   13	CIN			as above
   12	BINV		as above
11: 9	PAC			CHAN		Pin Action Control
    8	LSL			CHAN		Link Service Latch negation control
 7: 6	PSC			CHAN		Pin State Control
 5: 3	FLC			CHAN		FLag Control
    2	CIR			CHAN		Channel Interrupt Request
 1: 0	DECEND		as above
 
Format 3
--------

31:30	10
29:26	BCC			SEQ			Branch Condition Code
   25	FLS			SEQ			FLuSh control
24:16	BAF			SEQ			Branch Address Field
15:12	TBS			CHAN		Time Base Select and pin control
11: 9	PAC			as above
    8	BCF			SEQ			Branch Condition Field control
 7: 6	PSC			as above
 5: 3	FLC			as above
    2	CIR			as above
 1: 0	MTSR		CHAN		Match Transition detect Service Request inhibit

Format 4
--------

31:29	110
   28   RW			as above
27:26	NMA			SEQ			Next uPC Address mode control
   25	FLS			as above
24:23	BAF			as above
15:13	FLC			as above
   12	LSL			as above
11: 9	IOM			as above
 8: 2	AID			as above
 1: 0	DECEND		as above

Format 5
--------

31:29	111
28:25	T1ABS		as above
24:21	T3ABD		as above
20:19	SHF			as above
   18	SRC			as above
   17	CCL			as above
16: 9	T1BBI		AU 			B Bus Immediate data ( in T1 )
    8	LSL			as above
 7: 6	xx			don't care
 5: 3	FLC			as above
    2	CIR			as above
 1: 0	DECEND		as above

*/

#define	FORM1			0x01
#define	FORM2			0x02
#define	FORM3			0x04
#define	FORM4			0x08
#define	FORM5			0x10


/****************************************************************************/ 
/*						RAM related fields									*/
/****************************************************************************/ 

/*
*	The RW field controls whether a RAM operation reads or writes
*	to the RAM
*/
 
#define	rw_WIDTH		1
#define	RW_READ			0		/* RAM -> REG */
#define	RW_WRITE		1		/* REG -> RAM */
#define	RW_NIL			1

/*
*	The IOM field controls the register ( P or DIOB ) and the addressing
*	mode ( channel relative, DIOB indirect or direct ).
*	When using DIOB then DIOB acts as a byte address but the lowest bit
*	is ignored.
*	Channels 0 - 13 can not access their 7th and 8th parameter location
*	using channel-relative addressing.
*/

#define	iom_WIDTH		3
#define	IOM_P_PRM		0		/* P register, channel relative				*/
#define	IOM_P_DIOB		1		/* P register, DIOB indirect				*/
#define	IOM_P_AID		2		/* P register, direct address				*/
#define	IOM_DIOB_PRM	4		/* DIOB register, channel relative			*/
#define	IOM_DIOB_DIOB	5		/* DIOB register, DIOB indirect				*/
#define	IOM_DIOB_AID	6		/* DIOB register, direct address			*/
#define	IOM_NIL			7

/*
*	The AID field contains the absolute RAM address or the channel relative
*	address. In each case it is a *word* address.
*/

#define	aid_WIDTH		7
#define	AID_NIL			0x7f

/****************************************************************************/ 
/*					Channel control related fields							*/
/****************************************************************************/ 

/*
*	The TBS field controls the pin direction and selects the timebases
*/

#define	tbs_WIDTH		4
#define	TBS_IN11		0		/* Pin input, capture TCR1, match TCR1 		*/
#define	TBS_IN12		1		/* Pin input, capture TCR1, match TCR2		*/
#define	TBS_IN21		2		/* Pin input, capture TCR2, match TCR1		*/
#define	TBS_IN22		3		/* Pin input, capture TCR2, match TCR2		*/
#define	TBS_OUT11		4		/* Pin output, capture TCR1, match TCR1		*/
#define	TBS_OUT12		5		/* Pin output, capture TCR1, match TCR2		*/
#define	TBS_OUT21		6		/* Pin output, capture TCR2, match TCR1		*/
#define	TBS_OUT22		7		/* Pin output, capture TCR2, match TCR2		*/
#define	TBS_NIL			15		/* No change on previous setting			*/

/*
*	The PAC controls the pin state change related actions. The output 
*	state change occurs when match detection happens.
*/

#define	pac_WIDTH		3
#define	PAC_NO_CHANGE	0		/* Out: no change		In: no detect		*/
#define	PAC_HIGH		1		/* Out: set to 1		In: detect 0 -> 1	*/
#define	PAC_LOW			2		/* Out: set to 0		In: detect 1 -> 0	*/
#define	PAC_TOGGLE		3		/* Out: toggle pin		In: detect any edge	*/
#define	PAC_NIL			7

/*
*	The PSC field can directly force an output pin to a state.
*	This will only happen in the next microcycle. 
*	If PSC==0 and PAC is assigned in the same insn, the pin gets the new PAC.
*/

#define	psc_WIDTH		2
#define	PSC_PAC			0		/* Pin action controlled by PAC				*/
#define	PSC_HIGH		1		/* Force pin to 1							*/
#define	PSC_LOW			2		/* Force pin to 0							*/
#define	PSC_NIL			3

/*
*	The CCM field allows simultaneous configuration of PSC, PAC and TBS.
*	The relevant field in the instruction are ignored and 
*
*		PSC <= P[1:0];
*		PAC <= P[4:2];
*		TBS <= P[9:5];
*
*	is executed.
*/

#define	ccm_WIDTH		1
#define	CCM_USE			0		/* Use P as {TBS,PAC,PSC} source			*/
#define	CCM_NIL			1		/* Don't use P as config source 			*/

/*
*	The FLC field is the channel flag control. The two user flags can be set 
*	or reset using this field.
*/

#define	flc_WIDTH		3
#define	FLC_SET_F0		1		/* Flag 0 is set							*/
#define	FLC_SET_F1		3		/* Flag 1 is set							*/
#define	FLC_CLR_F0		0		/* Flag 0 is cleared						*/
#define	FLC_CLR_F1		2		/* Flag 1 is cleared						*/
#define	FLC_NIL			7		/* No change on flags						*/

/*
*	The MTSR field controls whether a match or transition can set
*	the relevant service request latch
*/

#define	mtsr_WIDTH		2
#define	MTSR_ENA		0		/* Enable service request generation		*/
#define	MTSR_DIS		1		/* Disable service request generation		*/
#define	MTSR_NIL		3		/* No change on ena/dis state				*/

/*
*	This field negates (clears, not inverts) the Transition Detect Latch
*/

#define	tdl_WIDTH		1
#define	TDL_NEGATE		0		/* Negate (clear) TDL						*/
#define	TDL_NIL			1		/* Does not change TDL						*/

/*
*	This field negates (clears, not inverts) the Match Recognition Latch
*/

#define	mrl_WIDTH		1
#define	MRL_NEGATE		0		/* Negate (clear) MRL						*/
#define	MRL_NIL			1		/* Does not change MRL						*/

/*
*	This field negates (clears, not inverts) the Link Service Latch
*/

#define	lsl_WIDTH		1
#define	LSL_NEGATE		0		/* Negate (clear) LSL						*/
#define	LSL_NIL			1		/* Does not change LSL						*/

/*
*	The ERW field can enable a write from the ERT to the MER
*	in the next microcycle. The ERT (accessible through the AU) should
*	contain a valid time value.
*/

#define	erw_WIDTH		1
#define	ERW_WRITE		0		/* Write ERT to MER in nect microcycle		*/
#define	ERW_NIL			1		/* Leave MER alone							*/

/*
*	The CIR field can cause the TPU to request a channel interrupt
*	from the CPU core.
*/

#define	cir_WIDTH		1
#define	CIR_REQ			0		/* Request for CPU interruption				*/
#define	CIR_NIL			1		/* Leave the CPU alone						*/

/****************************************************************************/ 
/*						Sequencing related fields							*/
/****************************************************************************/ 

/*
*	The DECEND field controls the decrementer and also can signal the
*	end of a state. Note, the decrementer has its effect when it is
*	actually 0, but it is still decremented so after it did its work, 
*	it is set to 15. Writing a 0 to it is equal to 16.
*/

#define	decend_WIDTH	2
#define	DECEND_RET		0		/* Derc <= Decr-1; if ( ! Decr ) uPC <= RAR	*/
#define	DECEND_REP		1		/* Decr <= Decr-1; uPC <= uPC + ! Decr;		*/
#define	DECEND_END		2		/* Finish current state						*/
#define	DECEND_NIL		3

/*
*	The NMA field controls the uPC 
*/

#define	nma_WIDTH		2
#define	NMA_GOTO		0		/* uPC <= BAF								*/
#define	NMA_CALL		1		/* uPC <= BAF; RAR <= FLS ? uPC+2 : uPC+1	*/ 
#define	NMA_RET			2		/* uPC <= RAR								*/
#define	NMA_NIL			3		/* Nothing, that is uPC <= uPC+1			*/

/*
*	The BCC field defines the condition on conditional branch.
*/

#define	bcc_WIDTH		4
#define	BCC_LESSTHAN	0		/* N ^ V - signed less than					*/
#define	BCC_LOWSAME		1		/* C | Z - unsigned lower or same			*/
#define	BCC_OVF			2		/* V     - overflow							*/
#define	BCC_NEG			3		/* N 	 - negative							*/
#define	BCC_CY			4		/* C	 - carry, unsigned higher			*/
#define	BCC_ZERO		5		/* Z	 - zero, equal						*/
#define	BCC_FLAG1		6		/* Channel flag 1							*/
#define	BCC_FLAG0		7		/* Channel flag 0							*/
#define	BCC_TDL			8		/* Transition Detect Latch					*/
#define	BCC_MRL			9		/* Match Recognition Latch					*/
#define	BCC_LSL			10		/* Link Service Latch						*/
#define	BCC_HSQ1		11		/* Host Sequence Bit 1						*/
#define	BCC_HSQ0		12		/* Host Sequence Bit 0						*/
#define	BCC_PSL			13		/* Pin State Latch							*/
#define	BCC_NEVER		15		/* Constant '0'								*/
#define	BCC_NIL			15

/*
*	The BCF field controls if the branch happens when the condition is
*	set or when it is not.
*/

#define	bcf_WIDTH		1
#define	BCF_CLEAR		0		/* Branch if condition bit is not set		*/
#define	BCF_SET			1		/* Branch if condition is set				*/
#define	BCF_NIL			1

/*
*	The FLS controls the flush of the prefetch queue on jumps
*/

#define	fls_WIDTH		1
#define	FLS_FLUSH		0		/* Flush prefetched insn, exec NOP in slot	*/
#define	FLS_NIL			1		/* No flush, exec prefetched insns in slot	*/

/*
*	Branch address (in longwords)
*/

#define	baf_WIDTH		9
#define	BAF_NIL			0x1ff

/****************************************************************************/ 
/*						AU related fields									*/
/****************************************************************************/ 

/*
*	T1ABS controls the A bus source for AU operations
*/

#define	t1abs_WIDTH		4
#define	T1ABS_P_LOW		0		/* {  8'b0,   P[  7:0 ] }					*/
#define	T1ABS_P_HIGH	1		/* {  8'b0,   P[ 15:8 ] }					*/
#define	T1ABS_DEC		2		/* { 12'b0, DEC[  3:0 ] }					*/
#define	T1ABS_CHAN		3		/* { 12'b0, CHAN[ 3:0 ] }					*/
#define	T1ABS_ZERO8		7		/* { 16'b0 }								*/
#define	T1ABS_ZERO_MER	4		/* { 16'b0 } and ERT <= MER; 	<***		*/
#define	T1ABS_P			8		/* P										*/
#define	T1ABS_A			9		/* A										*/
#define	T1ABS_SR		10		/* SR										*/
#define	T1ABS_DIOB		11		/* DIOB										*/
#define	T1ABS_TCR1		12		/* TCR1										*/
#define	T1ABS_TCR2		13		/* TCR2										*/
#define	T1ABS_ERT		14		/* ERT										*/
#define	T1ABS_ZERO16	15		/* 16'b0									*/
#define	T1ABS_NIL		15

/*
*	T1BBS controls the B bus source for AU operations
*	The B bus is special because some extra HW handles some special cases:
*
*	- If the decrementor is decrementing and the shift right is specified
*	  (operator :=>>) and shift is specified ('shift' subinsn) then B source
*	  is either the selected source or 16'b0 depending on the least 
*	  significant bit of the SR.
*
*	- If the B bus source is T1BBS_ZERO and Cin = 0 and Binv = 0 then
*	  the B bus will be 0x8000.
*/

#define	t1bbs_WIDTH		3
#define	T1BBS_P			0		/* P										*/
#define	T1BBS_A			1		/* A										*/
#define	T1BBS_SR		2		/* SR										*/
#define	T1BBS_DIOB		3		/* DIOB										*/
#define	T1BBS_ZERO		7		/* 16'b0									*/
#define	T1BBS_NIL		7

/*
*	B bus invert control
*/

#define	binv_WIDTH		1
#define	BINV_NEGATE		0		/* Invert all B bus bits ( ~ B[15:0] )		*/
#define	BINV_PONATE		1		/* Don't invert all B bus bits ( B[15:0] )	*/
#define	BINV_NIL		1

/*
*	Carry in control
*
*/

#define	cin_WIDTH		1
#define	CIN_CARRY		0		/* Carry in for the AU is 1					*/
#define	CIN_NOCARRY		1		/* Carry in for the AU is 0					*/
#define	CIN_NIL			1

/*
*	Condition code latch control
*/

#define	ccl_WIDTH		1
#define	CCL_LATCH		0		/* Latch AU cond codes (Z,C,N,V) in T2		*/
#define	CCL_NOLATCH		1		/* Do not latch condition codes				*/
#define	CCL_NIL			1

/*
*	AU shifter control. 
*	One tricky thing: when shifting to the right, if the shift register is
*	also shifting, then Cout is shifted in, otherwise 0.
*	
*/

#define	shf_WIDTH		2
#define	SHF_LEFT		0		/* CY <= AU[15]; AB <= { AU[14:1], 1'b0 }	*/
#define	SHF_RIGHT		1		/* CY<=AU[0]; AB<={SRC?Cout:1'b0, AU[15:1]} */
#define	SHF_ROTR		2		/* AB <= { AU[0], AU[15:1] }; CY <= AU[0]	*/
#define	SHF_NOSHIFT		3		/* CY <= Cout								*/
#define	SHF_NIL			3

/*
*	Shift register control (always shifts to the right)
*	If SHF == SHF_RIGHT then the input bit is AB[0], otherwise 1'b0
*/

#define	src_WIDTH		1
#define	SRC_SHIFT		0		/* SR is shifting right						*/
#define	SRC_NOSHIFT		1		/* SR is not shifting						*/
#define	SRC_NIL			1

/*
*	AU destination control
*/

#define	t3abd_WIDTH		4
#define	T3ABD_A			0		/* A[15:0] <= AB[15:0]						*/
#define	T3ABD_SR		1		/* SR[15:0] <= AB[15:0]						*/
#define	T3ABD_ERT		2		/* ERT[15:0] <= AB[15:0]					*/
#define	T3ABD_DIOB		3		/* DIOB[15:0] <= AB[15:0]					*/
#define	T3ABD_P_HIGH	4		/* P[15:0] <= { AB[7:0], P[7:0] }			*/
#define	T3ABD_P_LOW		6		/* P[15:0] <= { P[15:8], AB[7:0] }			*/
#define	T3ABD_P			7		/* P[15:0] <= AB[15:0]						*/
#define	T3ABD_LINK		8		/* LINK[3:0] <= AB[7:4]						*/
#define	T3ABD_CHAN		9		/* CHAN[3:0] <= AB[7:4]						*/
#define	T3ABD_DEC		10		/* DEC[3:0] <= AB[3:0]						*/
#define	T3ABD_CHANDEC	11		/* { CHAN[3:0], DEC[3:0] } <= AB[7:0]		*/
#define	T3ABD_TCR1		12		/* TCR1[15:0] <= AB[15:0]					*/
#define	T3ABD_TCR2		13		/* TCR2[15:0] <= AB[15:0]					*/
#define	T3ABD_NIL		15

/*
*	B bus immediate data
*/

#define	t1bbi_WIDTH		8
#define	T1BBI_NIL		0xff

/****************************************************************************/
/*				Function prototypes and function macros						*/
/****************************************************************************/

void	SetField( int form, FIELD *field, int val, char *name );
void	SetAID( int val1, int val2 );
void	SetBAF( IDENT *val );
void	SetBCF( int val );
void	SetNOP( void );
void	SetRMER( void );
void	SetT1BBI( int n );
void	TestConstant( int data, int expect );
void	SetBBSConst( int val );
void	SetT1ABS( int val );

#define	SetRW( val ) \
	SetField( FORM1 | FORM4, &insn.rw, val, "RW" )

#define	SetT3ABD( val ) \
	SetField( FORM1 | FORM2 | FORM5, &insn.t3abd, val, "T1ABD" )

#define	SetSHF( val ) \
	SetField( FORM1 | FORM2 | FORM5, &insn.shf, val, "SHF" )

#define	SetSRC( val ) \
	SetField( FORM1 | FORM5, &insn.src, val, "SRC" )

#define	SetCCL( val ) \
	SetField( FORM1 | FORM5, &insn.ccl, val, "CCL" )

#define	SetT1BBS( val ) \
	SetField( FORM1 | FORM2, &insn.t1bbs, val, "T1BBS" )

#define	SetCIN( val ) \
	SetField( FORM1 | FORM2, &insn.cin, val, "CIN" )

#define	SetBINV( val ) \
	SetField( FORM1 | FORM2, &insn.binv, val, "BINV" )

#define	SetIOM( val ) \
	SetField( FORM1 | FORM4, &insn.iom, val, "IOM" )

#define	SetDecEnd( val ) \
	SetField( FORM1 | FORM2 | FORM4 | FORM5, &insn.decend, val, "DECEND" )

#define	SetERW() \
	SetField( FORM2, &insn.erw, ERW_WRITE, "ERW" )

#define	SetTDL() \
	SetField( FORM2, &insn.tdl, TDL_NEGATE, "TDL" )

#define	SetMRL() \
	SetField( FORM2, &insn.mrl, MRL_NEGATE, "MRL" )

#define	SetLSL() \
	SetField( FORM2 | FORM4, &insn.lsl, LSL_NEGATE, "LSL" )

#define	SetCIR() \
	SetField( FORM2 | FORM5, &insn.cir, CIR_REQ, "CIR" )

#define	SetFLS( val ) \
	SetField( FORM3 | FORM4, &insn.fls, val, "FLS" )

#define	SetNMA( val ) \
	SetField( FORM4, &insn.nma, val, "NMA" )

#define	SetBCC( val ) \
	SetField( FORM3, &insn.bcc, val, "BCC" )

#define	SetFLC( val ) \
	SetField( FORM2 | FORM3 | FORM4 | FORM5 , &insn.flc, val, "FLC" )

#define	SetPSC( val ) \
	SetField( FORM2 | FORM3, &insn.psc, val, "PSC" )

#define	SetTBS( val ) \
	SetField( FORM3, &insn.tbs, val, "TBS" )

#define	SetCCM() \
	SetField( FORM3, &insn.ccm, CCM_USE, "CCM" )

#define	SetMTSR( val ) \
	SetField( FORM3, &insn.mtsr, val, "MTSR" )

#define	SetPAC( val ) \
	SetField( FORM2 | FORM3, &insn.pac, val, "PAC" )

#endif
