/****************************************************************************/
/*			  Motorola Time Processor Unit microcode assembler 				*/
/****************************************************************************/
/*																			*/
/*	Copyright 1998, (c) Bendor Research Pty. Ltd.							*/
/*																			*/
/*	This file is under the GNU General Public License.						*/
/*	See LICENSE for details.												*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	This is the main of the assembler.										*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	Changelog:																*/
/*																			*/
/*	30/04/98	Zoltan Kocsi	Alpha version								*/
/*	20/04/98	Zoltan Kocsi	Fixed an unitialised variable bug.			*/
/*																			*/
/****************************************************************************/

#include "tpuasm.h"

/****************************************************************************/
/*						Public data											*/
/****************************************************************************/

WORD	tpuram[ MAXRAM ];			/* Image of the TPU RAM	(insns)			*/
ENTRY	entries[ MAXENT ];			/* Image of the entries					*/

int		lineno;						/* Current line number					*/
int		src_lineno;					/* Current source line number			*/
char	*sourcename;				/* Current source name					*/


int		tab_size = 8;				/* Size of the tab chars				*/
int		max_line = 0x7fffffff;		/* Max length of src lines listed		*/

Bool	error = False;				/* Error occured						*/
Bool	lst_macro = False;			/* Flag: list macro expansions			*/

/****************************************************************************/
/*						Local prototypes									*/
/****************************************************************************/

static	Bool	FindArg( int *pos, int argc, char **argv );
static	int		SrecName( int now, int argc, char **argv );
static	int		ListName( int now, int argc, char **argv );
static	int		CIncName( int now, int argc, char **argv );
static	int		ListMacro( int now, int argc, char **argv );
static	int		InclDir( int now, int argc, char **argv );
static	int		DefType( int now, int argc, char **argv );
static	int		TabSize( int now, int argc, char **argv );
static	int		ListWidth( int now, int argc, char **argv );
static	int		SrecWidth( int now, int argc, char **argv );
static	int		SrecType( int now, int argc, char **argv );
static	int		SrecBase( int now, int argc, char **argv );
static	int		Altern( int now, int argc, char **argv );
static	int		Version( int now, int argc, char **argv );
static	void	Usage( char *name );
static	void	CreateHexFile( void );
static	void	CreateCFile( void );

/****************************************************************************/
/*						Local data											*/
/****************************************************************************/

static	int		srec_width = 28;		/* Srec width						*/
static	int		srec_type  = 1;			/* Srec type						*/
static	int		srec_base  = 0;			/* Srec type						*/
static	char	*srecname  = NULL;		/* Name of the srec file			*/
static	char	*cincname  = NULL;		/* Name of the C include file		*/
static	char	*listname  = NULL;		/* Name of the list file			*/
static	Bool	altern	   = False;		/* New type C interface				*/

static	const	struct	{

	char	*name;
	int		(*func)( int, int, char ** );
	
} args[] = {

	{ "-s",				SrecName	},
	{ "--srecfile",		SrecName	},
	{ "-l",				ListName	},
	{ "--listfile",		ListName	},
	{ "-c",				CIncName	},
	{ "--cincfile",		CIncName	},
	{ "-m",				ListMacro	},
	{ "--listmacro",	ListMacro	},
	{ "-I",				InclDir		},
	{ "--include",		InclDir		},
	{ "-M",				DefType		},
	{ "--memsize",		DefType		},
	{ "-t",				TabSize		},
	{ "--tabsize",		TabSize		},
	{ "-w",				ListWidth	},
	{ "--listwidth",	ListWidth	},
	{ "-W",				SrecWidth	},
	{ "--srecwidth",	SrecWidth	},
	{ "-S",				SrecType	},
	{ "--srectype",		SrecType	},
	{ "-B",				SrecBase	},
	{ "--srecbase",		SrecBase	},
	{ "-V",				Version		},
	{ "--version",		Version		},
	{ "-a",				Altern		},
	{ "--alternative",	Altern		},
	{ NULL, 			NULL		}
};

static	const	char	usage[] = 
#include "usage.h"
;

/****************************************************************************/
/*						The assembler, as such :-)							*/
/****************************************************************************/

/*
*	TPU assembler
*/

int		main( int argc, char *argv[] )
{
int		i;
FILE	*infile;

	/* Init a few things */
	
	for ( i = 0 ; i < MAXRAM ; i++ ) {
	
		tpuram[ i ].valid = Unknown;
		tpuram[ i ].line  = 0x7fffffff;
	}
	
	for ( i = 0 ; i < MAXENT ; i++ ) {
	
		entries[ i ].valid = Unknown;
	}
	
	AddIncludePath( "." );

	/* Process args */
	
	cif_init (argv[0]);

	if ( argc < 2 ) Usage( argv[ 0 ] );
	for ( i = 1 ; FindArg( &i, argc, argv ) ; );

	/* Prepare other stuff */
	
	InsnInitAll();
	EntryInit();
	ListfileInit();
	
	/* Start processing the input */
	
	for ( ; i < argc ; i++ ) {
	
		if ( ! strcmp( argv[ i ], "-" ) ) {
		
			sourcename = "-";
			SrcOpen( True, "<stdin>", stdin );
		}
		else if ( ! ( infile = fopen( argv[ i ], "r" ) ) ) {
		
			FatalError(( E_FOPENE, argv[ i ] ));
		}
		else {
		
			sourcename = argv[ i ];
			SrcOpen( True, argv[ i ], infile );
		}
		
		yyparse();
	}
	
	/* Finish all insn and entry dependencies if everything went well */
	
	if ( ! error ) {
	
		FixupInsn();
		FixupEntry();
	}
	
	/* Output is not generated unless the assembly went error free */
	
	if ( ! error ) {
		
		CreateListFile( listname );
		CreateHexFile();
		CreateCFile();
	}
	
	return( error ? EXIT_FAILURE : EXIT_SUCCESS );
}

/****************************************************************************/
/*						Local functions										*/
/****************************************************************************/

/*
*	Process command line options
*
*/

static	Bool	FindArg( int *pos, int argc, char **argv )
{
int		j;

	for ( j = 0 ; args[ j ].name ; j++ ) {
	
		if ( ! strcmp( argv[ *pos ], args[ j ].name ) ) {
		
			*pos = (*args[ j ].func)( *pos, argc, argv ) + 1;
			return( True );
		}
	}
	
	return( False );
}

/*
*	Various comman line option processing functions.
*/

static	int		SrecName( int now, int argc, char **argv )
{
	if ( now >= argc - 2 ) Usage( argv[ 0 ] );
	srecname = argv[ ++now ];
	return( now );
}

static	int		ListName( int now, int argc, char **argv )
{
	if ( now >= argc - 2 ) Usage( argv[ 0 ] );
	listname = argv[ ++now ];
	return( now );
}

static	int		CIncName( int now, int argc, char **argv )
{
	if ( now >= argc - 2 ) Usage( argv[ 0 ] );
	cincname = argv[ ++now ];
	return( now );
}

static	int		ListMacro( int now, int argc, char **argv )
{
	lst_macro = True;
	return( now );
}

static	int		InclDir( int now, int argc, char **argv )
{
	if ( now >= argc - 2 ) Usage( argv[ 0 ] );
	AddIncludePath( argv[ ++now ] );
	return( now );
}

static	int		DefType( int now, int argc, char **argv )
{
int		n;

	if ( now >= argc - 2 ) Usage( argv[ 0 ] );
	if ( sscanf( argv[ ++now ], "%d", &n ) != 1 ) Usage( argv[ 0 ] );
	if ( n != 256 && n != 512 ) Usage( argv[ 0 ] );
	maxmem = n;
	cif_memend (n);
	return( now );
}

static	int		TabSize( int now, int argc, char **argv )
{
int		n;

	if ( now >= argc - 2 ) Usage( argv[ 0 ] );
	if ( sscanf( argv[ ++now ], "%d", &n ) != 1 ) Usage( argv[ 0 ] );
	if ( n < 2 || n > 16 ) Usage( argv[ 0 ] );
	tab_size = n;
	return( now );
}

static	int		ListWidth( int now, int argc, char **argv )
{
int		n;

	if ( now >= argc - 2 ) Usage( argv[ 0 ] );
	if ( sscanf( argv[ ++now ], "%d", &n ) != 1 ) Usage( argv[ 0 ] );
	if ( n < 2 ) Usage( argv[ 0 ] );
	max_line = n;
	return( now );
}

static	int		SrecWidth( int now, int argc, char **argv )
{
int		n;

	if ( now >= argc - 2 ) Usage( argv[ 0 ] );
	if ( sscanf( argv[ ++now ], "%d", &n ) != 1 ) Usage( argv[ 0 ] );
	if ( n < 1 || n > 128 ) Usage( argv[ 0 ] );
	srec_width = n;
	return( now );
}

static	int		SrecType( int now, int argc, char **argv )
{
int		n;

	if ( now >= argc - 2 ) Usage( argv[ 0 ] );
	if ( sscanf( argv[ ++now ], "%d", &n ) != 1 ) Usage( argv[ 0 ] );
	if ( n < 1 || n > 3 ) Usage( argv[ 0 ] );
	srec_type = n;
	return( now );
}

static	int		SrecBase( int now, int argc, char **argv )
{
int		n;

	if ( now >= argc - 2 ) Usage( argv[ 0 ] );
	now++;
	
	if ( argv[ now ][ 0 ] == '0' &&
		 ( argv[ now ][ 1 ] == 'x' || argv[ now ][ 1 ] == 'X' ) ) {
	
		if ( sscanf( argv[ now ], "%x", &n ) != 1 ) Usage( argv[ 0 ] );
	}
	else {
		if ( sscanf( argv[ now ], "%d", &n ) != 1 ) Usage( argv[ 0 ] );
	}
		
	srec_base = n;
	return( now );
}

static	int		Version( int now, int argc, char **argv )
{
	printf( "%s\n", VERSION );
	exit( 0 );
	return( 0 );
}

/*
*	Sets alternative C interface
*/

static	int		Altern( int now, int argc, char **argv )
{
	altern = True;
	return( now );
}

/*
*	Prints a short usage to the user.
*/

static	void	Usage( char *name )
{
	fprintf( stderr, usage, name );
	exit( 1 );	
}

/*
*	This function  creates the S-record file
*/

static void CreateHexFile( void )
{
int		i;

	if ( ! srecname ) return;
	
	SrecOpen( srecname, srec_type, srec_width );
	
	for ( i = 0 ; i < maxmem ; i++ ) {
	
		if ( tpuram[ i ].valid == True ) {
		
			SrecAddress( srec_base + 4*i );
			SrecLong( tpuram[ i ].value );
		}
	}
	
	for ( i = 0 ; i < MAXENT ; i++ ) {
	
		if ( entries[ i ].valid == True ) {
		
			SrecAddress( srec_base + ( maxmem - 128 ) * 4 + 2 * i );
			SrecWord( entries[ i ].value );
		}
	}
	
	SrecClose();
}

/*
*	This function  creates the C include file
*/

static void CreateCFile( void )
{
FILE	*fp;
int		i, j;
ulong	tmp[ MAXRAM ];

	if ( ! cincname ) return;
	
	if ( ! strcmp( cincname, "-" ) ) {
	
		fp = stdout;
	}
	else {
	
		if ( ! ( fp = fopen( cincname, "w" ) ) ) {
	
			FatalError(( E_FOPENE, cincname ));
		}
	}
	
	for ( i = 0 ; i < MAXRAM ; i++ ) tmp[ i ] = 0xffffffff;
	
	for ( i = 0 ; i < maxmem ; i++ )
	
		if ( tpuram[ i ].valid == True )
		
			tmp[ i ] = tpuram[ i ].value;
	
	for ( i = 0, j = maxmem - 128 ; i < MAXENT ; i += 2, j++ ) {
	
		if ( entries[ i ].valid == True ) {
		
			tmp[ j ] &= 0x0000ffff;
			tmp[ j ] |= entries[ i ].value << 16;
		}
		
		if ( entries[ i+1 ].valid == True ) {
		
			tmp[ j ] &= 0xffff0000;
			tmp[ j ] |= entries[ i+1 ].value;
		}
	}
	
	if ( altern ) {
	
		cif_flush(sourcename, fp);
	}
	else {
	
		for ( i = 0 ; i < maxmem ; i++ ) {
	
			fprintf( fp, "0x%08x", (int) tmp[ i ] );
			if ( i != maxmem - 1 ) putc( ',', fp );
			putc( ( i & 3 ) == 3 ? '\n' : ' ', fp );
		}
	}
	
	if ( fp != stdout ) fclose( fp );
}
