%{

/********************************************************************/
/*		  Motorola Time Processor Unit microcode assembler 			*/
/********************************************************************/
/*																	*/
/*	Copyright 1998, (c) Bendor Research Pty. Ltd.					*/
/*																	*/
/*	This file is under the GNU General Public License.		 		*/
/*	See LICENSE for details.										*/
/*																	*/
/********************************************************************/
/*																	*/
/*	This file is the lexer for the assembler.						*/
/*																	*/
/********************************************************************/
/*																	*/
/*	Changelog:														*/
/*																	*/
/*	30/04/98	Zoltan Kocsi	1.0 Alpha version					*/	
/*																	*/
/********************************************************************/
		
#include "tpuasm.h"
#include "y.tab.h"

/****************************************************************************/
/*				Public data													*/
/****************************************************************************/

YYSTYPE	yylval;							/* Associated token value			*/
Bool	in_macro;						/* Currently processing a macro		*/

/****************************************************************************/
/*				Local types and defines										*/
/****************************************************************************/

typedef	struct s_input	{

	struct s_input	*next;				/* Link pointer	to previous source	*/
	int				linesave;			/* Previous src linenumber			*/
	Bool			in_macro;			/* Previous src in_macro value		*/
	YY_BUFFER_STATE	prevb;				/* Previous src flex buffer			*/
	char			*name;				/* Name of this source				*/
	FILE			*filep;				/* If file, the filepointer			*/
	
} SOURCE;								/* Source descriptor				*/

/****************************************************************************/
/*				Prototypes of local functions								*/
/****************************************************************************/

static	int		yywrap( void );
static	void	decimal( void );
static	void	hex( void );
static	void	binary( void );
static	void	create_string( void );
static	void	create_ident( void );
static	void	substitute_macro( void );

/****************************************************************************/
/*				Static data													*/
/****************************************************************************/

static	SOURCE	*current_input = NULL;

%}

	/********************************************************************/
	/*					Lexer definitions								*/
	/********************************************************************/
	
	/*
	*	The TPU assembler accepts Pascal style comments. (Apparently it
	*	was originally written by Pascal programmers, there are a lot
	*	of Pascal-ish things in the syntax.)
	*	So, comments are enclosed between either (* and *) or { and }.
	*	Nested comments are allowed if the nesting and the nested
	*	(how about nestee and nester :-) are of different kind.
	*/
	
%s	COMMENT1
%s	COMMENT2
	
%%

	/********************************************************************/
	/*					Process comments								*/
	/********************************************************************/

"(*"						{ Token( yytext, False ); BEGIN COMMENT1; }
<COMMENT1>"*"+")"			{ Token( yytext, False ); BEGIN 0; }
<COMMENT1>\n				{ Token( yytext, False ); }
<COMMENT1>[^*\n\r]*			{ Token( yytext, False ); }
<COMMENT1>"*"+[^*)\n\r]*	{ Token( yytext, False ); }
<COMMENT1>\r				{ ; }
<COMMENT1><<EOF>>			{ FatalError(( E_EOFCOM )); }

"{"							{ Token( yytext, False ); BEGIN COMMENT2; }
<COMMENT2>"}"				{ Token( yytext, False ); BEGIN 0; }
<COMMENT2>\n				{ Token( yytext, False ); }
<COMMENT2>[^}\n\r]*			{ Token( yytext, False ); }
<COMMENT2>\r				{ ; }
<COMMENT2><<EOF>>			{ FatalError(( E_EOFCOM )); }

	/********************************************************************/
	/*					Process whitespace								*/
	/********************************************************************/

[ \t]*				{ Token( yytext, False ); }
\n					{ Token( yytext, False ); }
\r					{ ; }

	/********************************************************************/
	/*				Single char operators								*/
	/********************************************************************/
	
[\#\+\-\!,;\.='*\(\):]	{ Token( yytext, False ); return( yytext[ 0 ] ); }

	/********************************************************************/
	/*				Multiple character operators						*/
	/********************************************************************/
	
:=			{ /* Assignment */ 
			  Token( yytext, False ); return( LET ); }

:=>>		{ /* Assignment and shift right */ 
			  Token( yytext, False ); return( LET_SR ); }

:=<<		{ /* Assignment and shift left */ 
			  Token( yytext, False ); return( LET_SL ); }

:=R>		{ /* Assignment and rotate right */ 
			  Token( yytext, False ); return( LET_RR ); }
			  
\-\>		{ /* Write to memory */
			  Token( yytext, False ); return( WRITE ); }
			 
\<\-		{ /* Read from memory */
			  Token( yytext, False ); return( READ ); }
			 
	/********************************************************************/
	/*						Subinstructions								*/
	/********************************************************************/


au			{ /* Arithmetic unit */ 
			  Token( yytext, False ); return( AU ); }

call		{ /* Call subroutine */ 
			  Token( yytext, False ); return( CALL ); }

chan		{ /* Channel control */ 
			  Token( yytext, False ); return( CHAN ); }

dec_return	{ /* Decrement and return */ 
			  Token( yytext, False ); return( DEC_RET ); }

end			{ /* End of state */ 
			  Token( yytext, False ); return( END ); }

goto		{ /* Unconditional branch */ 
			  Token( yytext, False ); return( GOTO ); }

if			{ /* Conditional branch */ 
			  Token( yytext, False ); return( IF ); }
	
then		{ /* Conditional stuffing a' la COBOL :-) */
			  Token( yytext, False ); return( THEN ); }

nop			{ /* No operation */ 
			  Token( yytext, False ); return( NOP ); }

ram			{ /* RAM operations */ 
			  Token( yytext, False ); return( RAM ); }

repeat		{ /* Repeat microinstruction */ 
			  Token( yytext, False ); return( REPEAT ); }

return		{ /* Return from subroutine */ 
			  Token( yytext, False ); return( RETURN ); }
			  
	/********************************************************************/
	/*						Arithmetical unit 							*/
	/********************************************************************/
	   
	/*
	*	Data source and destinations
	*/
	
a			{ /* Accumulator */ 
			  Token( yytext, False ); return( A ); }
			
sr			{ /* Shift register */
			  Token( yytext, False ); return( SR ); }
			
ert			{ /* Event register temporary */ 
			  Token( yytext, False ); return( ERT ); }
		
diob		{ /* Data input/output buffer register */ 
			  Token( yytext, False ); return( DIOB ); }

p_high		{ /* P register high (8 bit) */ 
			  Token( yytext, False ); return( PHIGH ); }

p_low		{ /* P register low (8 bit) */ 
			  Token( yytext, False ); return( PLOW ); }

p			{ /* P register as a word */ 
			  Token( yytext, False ); return( P ); }

link		{ /* Link register */ 
			  Token( yytext, False ); return( LINK ); }

chan_reg	{ /* Channel register (8 bit) */ 
			  Token( yytext, False ); return( CHANREG ); }

dec			{ /* Decrementer (8 bit) */ 
			  Token( yytext, False ); return( DEC ); }

chan_dec	{ /* Concetenated chan_reg and dec */ 
			  Token( yytext, False ); return( CHDEC ); }

tcr1		{ /* Time counter reg 1	*/ 
			  Token( yytext, False ); return( TCR1 ); }

tcr2		{ /* Time counter reg 2 */ 
			  Token( yytext, False ); return( TCR2 ); }

nil			{ /* No destination	*/ 
			  Token( yytext, False ); return( NIL ); }

	/*
	*	Arithmetical unit qualifiers
	*/
	
ccl			{ /* Condition codes latched at the end of the cycle */
			  Token( yytext, False ); return( CCL ); }
			
shift		{ /* Shift the SR to the right by one bit */
			  Token( yytext, False ); return( SHIFT ); }

read_mer	{ /* Read the Channel Match Event register into the ERT */
			  Token( yytext, False ); return( RD_MER ); }
			
	/*
	*	Special constants
	*/

max			{ /* This is a constant that is equal to 0x8000 */
			  Token( yytext, False ); yylval.val = 0x8000; return( CONST ); }
	
	/********************************************************************/
	/*						Flow control	 							*/
	/********************************************************************/


flush		{ /* Flush insn queue (prefetch reg) before jumping */
			  Token( yytext, False ); return( FLUSH ); }
		
no_flush	{ /* Do not flush, that is, execute the next insn before
				 jumping, just like most RISCs do */
			  Token( yytext, False ); return( NO_FLUSH ); }


	/********************************************************************/
	/*						Channel control 							*/
	/********************************************************************/

	/*
	*	Flag operations
	*/
	
set			{ /* Sets a flag */
			  Token( yytext, False ); return( SET ); }
			  
clear		{ /* Clears a flag */
			  Token( yytext, False ); return( CLEAR ); }

flag0		{ /* User settable/clearable/testable flag 0 */
			  Token( yytext, False ); return( FLAG0 ); }

flag1		{ /* User settable/clearable/testable flag 1 */
			  Token( yytext, False ); return( FLAG1 ); }

	/*
	*	PAC (Pin Action Control)
	*/
	
pac			{ /* mi a fene az a pac */
			  Token( yytext, False ); return( PAC ); }
	
high		{ /* Set to high */
			  Token( yytext, False ); return( HIGH ); }

low			{ /* Set to low */
			  Token( yytext, False ); return( LOW ); }

no_change	{ /* No change on pin */
			  Token( yytext, False ); return( NO_CHANGE ); }
	
toggle		{ /* Change pin state */
			  Token( yytext, False ); return( TOGGLE ); }
	
low_high	{ /* Detect rising edge */
			  Token( yytext, False ); return( LOW_HIGH ); }

high_low	{ /* detect falling edge */
			  Token( yytext, False ); return( HIGH_LOW ); }
			  
no_detect	{ /* Do not monitor the pin */
			  Token( yytext, False ); return( NO_DETECT ); }
	
any_trans	{ /* Detect any transition on the pin */
			  Token( yytext, False ); return( ANY_TRANS ); }
			  
	/*
	*	PSC (Pin State Control)
	*/
	
pin			{ /* This represents the pin :-) */
			  Token( yytext, False ); return( PIN ); }

	/*
	*	TBS (Time Base Select)
	*/
	
tbs			{ /* Time base selector */
			  Token( yytext, False ); return( TBS ); }

in_m1_c1	{ /* Input channel, capture TCR1, macth TCR1 */
			  Token( yytext, False ); return( IN_M1_C1 ); }

in_m2_c1	{ /* Input channel, capture TCR1, macth TCR2 */
			  Token( yytext, False ); return( IN_M2_C1 ); }

in_m1_c2	{ /* Input channel, capture TCR2, macth TCR1 */
			  Token( yytext, False ); return( IN_M1_C2 ); }

in_m2_c2	{ /* Input channel, capture TCR2, macth TCR2 */
			  Token( yytext, False ); return( IN_M2_C2 ); }

out_m1_c1	{ /* Output channel, capture TCR1, macth TCR1 */
			  Token( yytext, False ); return( OUT_M1_C1 ); }

out_m2_c1	{ /* Output channel, capture TCR1, macth TCR2 */
			  Token( yytext, False ); return( OUT_M2_C1 ); }

out_m1_c2	{ /* Output channel, capture TCR2, macth TCR1 */
			  Token( yytext, False ); return( OUT_M1_C2 ); }

out_m2_c2	{ /* Output channel, capture TCR2, macth TCR2 */
			  Token( yytext, False ); return( OUT_M2_C2 ); }

	/*
	*	Other actions
	*/
	
write_mer	{ /* Write the Match Event Register */
			  Token( yytext, False ); return( WR_MER ); }
			  
neg_tdl		{ /* Negate Transition Detect Latch */
			  Token( yytext, False ); return( NEG_TDL ); }
			  
neg_mrl		{ /* Negate Match Recognition Latch */
			  Token( yytext, False ); return( NEG_MRL ); }
			  
neg_lsl		{ /* Negate Loutk Service Latch */
			  Token( yytext, False ); return( NEG_LSL ); }

cir			{ /* Assert channel interrupt request */
			  Token( yytext, False ); return( CIR ); }
			  
disable_mtsr { /* Disable service request */
			  Token( yytext, False ); return( DIS_MTSR ); }

enable_mtsr { /* Enable service request */
			  Token( yytext, False ); return( ENA_MTSR ); }
			  
config		{ /* Configure channel */
			  Token( yytext, False ); return( CONFIG ); }

	/********************************************************************/
	/*						Conditions		 							*/
	/********************************************************************/

	/*
	*	Condition codes
	*/
		
less_than	{ /* ASRC was less than BSRC (signed) */
			  Token( yytext, False ); return( LESS_THAN ); }
			  
low_same	{ /* ASRC was lower or same than BSRC (unsigned) */
			  Token( yytext, False ); return( LOW_SAME ); }
			 
v			{ /* Overflow */
			  Token( yytext, False ); return( OVF ); }

n			{ /* Negative */
			  Token( yytext, False ); return( NEG ); }
			  
c			{ /* Carry */
			  Token( yytext, False ); return( CY ); }
			  
z			{ /* Zero */
			  Token( yytext, False ); return( ZERO ); }
			  
tdl			{ /* Transition Detect Latch */
			  Token( yytext, False ); return( TDL ); }
			  
mrl			{ /* Match Recognition Latch */
			  Token( yytext, False ); return( MRL ); }
			  
lsl			{ /* Link Service Latch */
			  Token( yytext, False ); return( LSL ); }
			  
psl			{ /* Pin state Latch */
			  Token( yytext, False ); return( PSL ); }
			  
hsq0		{ /* Sequence bit 0 */
			  Token( yytext, False ); return( HSQ0 ); }
			  
hsq1		{ /* Sequence bit 1 */
			  Token( yytext, False ); return( HSQ1 ); }
			  
	/*
	*	Condition values
	*/
	
true		{ /* Same as '1' */
			  Token( yytext, False ); yylval.val = 1; return( CONST ); }
	
false		{ /* Same as '0' */
			  Token( yytext, False ); yylval.val = 0; return( CONST ); }
			  
			  
	/********************************************************************/
	/*					RAM access address codes						*/
	/********************************************************************/
			  
	/*
	*	Relative addressing, the memory accessed is the channel's own
	*	parameter area.
	*/
	
prm[0-7]	{ /* Parameter slot 0 - 7 */
			  Token( yytext, False ); yylval.val = yytext[3] - '0'; return( PRM ); }


	/*
	*	DIOB[7:1] addresses the RAM (DIOB[0] is ignored)
	*/
	
by_diob		{ /* Indirect address in DIOB */
			  Token( yytext, False ); return( BY_DIOB ); }
			  

	/********************************************************************/
	/*						Directives									*/
	/********************************************************************/
	
"%entry"		{ /* Entry point definition */
				  Token( yytext, False ); return( DENTR ); }
	
function		{ /* Function identifier for %entry */
				  Token( yytext, False ); return( FUNCTION ); }
				  
name			{ /* Name identifier for %entry */
				  Token( yytext, False ); return( NAME ); }

start_address	{ /* Entry address for %entry */
				  Token( yytext, False ); return( S_ADDR ); }
				  
enable_match	{ /* */
				  Token( yytext, False ); return( ENA_MATCH ); }
				  
disable_match	{ /* */
				  Token( yytext, False ); return( DIS_MATCH ); }

cond			{ /* Condition at which the entry applies */
				  Token( yytext, False ); return( COND ); }

hsr0			{ /* Condition: Host Service Request reg. 0 */
				  Token( yytext, False ); return( HSR0 ); }
				  
hsr1			{ /* Condition: Host Service Request reg. 1 */
				  Token( yytext, False ); return( HSR1 ); }
				  
lsr				{ /* Condition: Link Service Request */
				  Token( yytext, False ); return( LSR ); }
				  
"m/tsr"			{ /* Condition: Match/Transition */
				  Token( yytext, False ); return( MTSR ); }
x				{ /* Don't care value */
				  Token( yytext, False ); return( X ); }
				
"%include"		{ /* */
				  Token( yytext, False ); return( INCLUDE ); }
				  
"%macro"		{ /* */
				  Token( yytext, False ); return( MACRO_DEF ); }
				  
"%org"			{ /* */
				  Token( yytext, False ); return( ORG ); }
				  
"%page"			{ /* */
				  Token( yytext, False ); return( PAGE ); }
				  
"%type"			{ /* */
				  Token( yytext, False ); return( TYPE ); }
				  
tpu1			{ /* The only type known to the assembler */
				  Token( yytext, False ); return( TYPE_1 ); }
				  
256				{ /* Size of the TPU RAM */
				  Token( yytext, False ); return( TYPE_256 ); }
				  
512				{ /* Size of the TPU RAM */
				  Token( yytext, False ); return( TYPE_512 ); }
				  
	/********************************************************************/
	/*						Complex things								*/
	/********************************************************************/

[a-zA-Z][a-zA-Z0-9_\\]*		{ Token( yytext, False ); 
							  create_ident();
							  return( LABEL ); }

\@[a-zA-Z][a-zA-Z0-9_\\]*	{ substitute_macro(); }

\'[^']*\'						{ Token( yytext, False ); 
							  create_string(); 
							  return( STRING ); }
							  
[0-9]+						{ Token( yytext, False ); decimal(); return( CONST ); }
$[0-9a-fA-F]+				{ Token( yytext, False ); hex(); return( CONST ); }
%[01]+						{ Token( yytext, False ); binary(); return( CONST ); }

%%

/****************************************************************************/
/*				File and string inclusion handling							*/
/****************************************************************************/

/*
*	This function opens a file or a string for reading
*/

void	SrcOpen( Bool file, char *name, void *value )
{
SOURCE	*p;
int		i;
YY_BUFFER_STATE	buf;

	/* Check for recursive macros */
	
	if ( ! file ) {
	
		for ( p = current_input ; p ; p = p->next ) {
		
			if ( p->filep ) continue;
			
			for ( i = 0 ; name[ i ] && p->name[ i ] ; i++ )
			
				if ( toupper( name[ i ] ) != toupper( p->name[ i ] ) )
				
					break;
			
			if ( ! name[ i ] && ! p->name[ i ] ) {
			
				FatalError(( E_RECMAC, name ));
			}
		}
	}
	
	/* Create a new source entry */
	
	if ( ! ( p = malloc( sizeof( SOURCE ) ) ) ) {
	
		fprintf( stderr, "Out of memory !\n" );
		exit( 2 );
	}
	
	p->name = name;
	p->next  = current_input;
	p->prevb = YY_CURRENT_BUFFER;
	p->linesave = src_lineno;
	p->in_macro = in_macro;
	current_input = p;
	
	if ( file ) {
	
		if ( ! value ) {
		
			if ( ! ( value = fopen( name, "r" ) ) ) {
		
				fprintf( stderr, "Can't open %s\n !", name );
				exit( 2 );
			}
		}
		
		p->filep = (FILE *) value;
		yyin = (FILE *) value;
		buf = yy_create_buffer( (FILE *) value, YY_BUF_SIZE );
		yy_switch_to_buffer( buf );
		src_lineno = 1;
	}
	else {
		
		yy_scan_string( value );
		src_lineno = 0;
		p->filep = NULL;
	}
}

/*
*	This function is called by flex when it reaches the end of a file or macro
*/

static	int		yywrap( void )
{
SOURCE	*p;
YY_BUFFER_STATE	buf;

	p = current_input;
	current_input = p->next;
	
	yy_delete_buffer( YY_CURRENT_BUFFER );
	buf = p->prevb;
	src_lineno = p->linesave;
	
	if ( p->filep ) { 
	
		fclose( p->filep );
		src_lineno--;
	}
	
	in_macro = p->in_macro;
	free( p );
	
	if ( current_input ) {
	
		yy_switch_to_buffer( buf );
		sourcename = current_input->name;
		return( 0 );
	}
	else
		return( 1 );
}

/*
*	This function generates the current list of inclusions 
*	to the stderr
*/

void	ListCurrent( void )
{
SOURCE	*p;
Bool	prevfil;

	if ( ! ( p = current_input ) ) return;;
	
	if ( p->filep ) {
		
		fprintf( stderr, E_LINETX, p->name, src_lineno );
		prevfil = True;
	}
	else {
	
		fprintf( stderr, E_MACRTX, p->name );
		prevfil = False;
	}
	
	for ( p  = current_input ; p->next ; p = p->next ) {

		fprintf( stderr, " %s", prevfil ? E_INCLTX : E_INVOTX );
		
		if ( p->next->filep ) {
		
			fprintf( stderr, E_LINETX, p->next->name, p->linesave );
			prevfil = True;
		}
		else {
	
			fprintf( stderr, E_MACRTX, p->next->name );
			prevfil = False;
		}
	}
	
	putc( '\n', stderr );
}

/****************************************************************************/
/*						Local functions										*/
/****************************************************************************/


/*
*	Calculates the value of a decimal number
*/

static	void	decimal( void )
{
	sscanf( yytext, "%d", &yylval.val );
}
	
/*
*	Calculates the value of a hex number
*/

static	void	hex( void )
{
	sscanf( yytext + 1, "%x", &yylval.val );
}

/*
*	Calculates the value of a binary number
*/

static	void	binary( void )
{
char	*p;

	yylval.val = 0;
	
	for ( p = yytext + 1 ; *p ; p++ ) {
	
		yylval.val <<= 1;
		if ( *p == 1 ) yylval.val |= 1;
	}
}

/*
*	Creates a string, omits the enclosing '-s
*/

static	void	create_string( void )
{
	if ( ! ( yylval.string = malloc( strlen( yytext ) + 1 ) ) ) {
							  
	  	FatalError(( E_OUTMEM ));
	}
	
	yytext[ strlen( yytext ) - 1 ] = 0;
	strcpy( yylval.string, yytext+1 );
}

static	void	create_ident( void )
{
	yylval.ident = IdentLookup( yytext );
}

static	void	substitute_macro( void )
{
MACRO	*p;

	Token( yytext, True );
	
	if ( ( p = MacroSearch( yytext+1 ) ) ) {
	
		SrcOpen( False, p->name, p->value );
		in_macro = True;
	}
	else {
	
		IssueError(( E_UNDMAC, yytext+1 ));
	}
}
