/****************************************************************************/
/*			  Motorola Time Processor Unit microcode assembler 				*/
/****************************************************************************/
/*																			*/
/*	Copyright 1998, (c) Bendor Research Pty. Ltd.							*/
/*																			*/
/*	This file is under the GNU General Public License.						*/
/*	See LICENSE for details.												*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	This file contains the common definitions used in the assembler.		*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	Changelog:																*/
/*																			*/
/*	30/04/98	Zoltan Kocsi	Alpha version								*/
/*	20/05/98	Zoltan Kocsi	Made typedefs conditional					*/
/*																			*/
/****************************************************************************/

#ifndef	__TPUASM_H__
#define	__TPUASM_H__

# include <ctype.h>
# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <errno.h>

/****************************************************************************/
/*						Common constants and types							*/
/****************************************************************************/

#define	VERSION		"1.0.0"

#ifndef __DONT_ULONG__
typedef	unsigned long	ulong;
#endif
#ifndef __DONT_USHORT__
typedef	unsigned short	ushort;
#endif
#ifndef __DONT_UCHAR__
typedef	unsigned char	uchar;
#endif

#define	MAXRAM		512				/* Max. # of TPU insn RAM words			*/
#define	MAXENT		256				/* Max. # of entries					*/
#define	MAXLINE		128				/* Max length of src line in listfile	*/

/*
*	DO NOT CHANGE FALSE AND TRUE !!!
*	Quite a few things assume that they are 0 and 1, respectively.
*/

typedef	enum {

	False = 0,						/* Boolean false						*/
	True = 1,						/* Boolean true							*/
	DontCare = 2,					/* Boolean don't care					*/
	Unknown = 3						/* Boolean unknown value				*/
	
} Bool;								/* Boolean 								*/

/****************************************************************************/
/*				Identifier related definitions								*/
/****************************************************************************/

typedef	struct	s_macr {

	struct	s_macr	*next;
	char			*name;
	char			*value;
	
} MACRO;


typedef	struct s_id	{

	struct	s_id	*next;			/* Link pointer							*/
	char			*name;			/* Name of identifier					*/
	int				id;				/* Assigned unique ID					*/
	int				value;			/* Actual value associated with label	*/
	int				is_set;			/* 'value' field is already set			*/
	
} IDENT;

IDENT	*IdentLookup( char *s );
IDENT	*IdentById( int id );
void	IdentSet( IDENT *ident, int val );

/****************************************************************************/
/*				Identifier related definitions								*/
/****************************************************************************/

typedef	enum {

	EmatchHsr0,
	EmatchHsr1,
	EmatchLsr,
	EmatchMtsr,
	EmatchPin,
	EmatchFlag0,
	EmatchLsrX,
	EmatchMtsrX,
	EmatchPinX,
	EmatchFlag0X
	
} Ematch;

/*
*	Be sure that this matches the Ematch definition
*/

#define	MATCHNAME "hsr0", "hsr1", "lsr", "m/tsr", "pin", "flag0", \
								  "lsr", "m/tsr", "pin", "flag0"

/****************************************************************************/
/*							File handling									*/
/****************************************************************************/

/*
*	This function opens a file or a string for reading
*/

void	SrcOpen( Bool file, char *name, void *value );

/****************************************************************************/
/*							Macro handling									*/
/****************************************************************************/

/*
*	Macro definition
*/

void	MacroDef( char *name, char *value );

/*
*	Macro lookup
*/

MACRO	*MacroSearch( char *name );

/****************************************************************************/
/*						S record generation									*/
/****************************************************************************/

/*
*	Initialises the SREC file. 
*/

void	SrecOpen( char *srecname, int type, int maxbytes );

/*
*	Sets the address for the next data item
*/

void	SrecAddress( ulong addr );

/*
*	Writes a 4 byte data entity
*/

void	SrecLong( ulong data );

/*
*	Writes a 2 byte data entity
*/

void	SrecWord( ushort data );

/*
*	Finishes the SREC file and closes it
*/

void	SrecFinish( void );
/****************************************************************************/
/*						Error handling										*/
/****************************************************************************/

#define	IssueError( s )				\
{									\
	sprintf s;					 	\
	CreateErrorEntry(); 			\
}

#define	IssueWarning( s )			\
{									\
	sprintf s;					 	\
	CreateWarningEntry(); 			\
}

#define	FatalError( s )				\
{ 									\
	sprintf s;					 	\
	Fatal();						\
}

#define	FixupError( sr, ln, s )		\
{ 									\
	sprintf s;			 			\
	CreateFixupError( sr, ln );		\
}

void	Fatal( void );
void	CreateErrorEntry( void );
void	CreateWarningEntry( void );
void	CreateFixupError( char *src, int lin );

/****************************************************************************/
/*				Instruction and entry encoding								*/
/****************************************************************************/

/*
*	Instruction subfield
*/

typedef	struct	{

	int		value;						/* Field value						*/
	Bool	is_set;						/* Field is set						*/
	
} FIELD;								/* Microinsn field					*/

/*
*	Definition of the various fields in various insns.
*	For details see insn.h
*/

typedef	struct {

	int		form;						/* Allowed insn format bitmask		*/
	Bool	error;						/* Insn is invalid due to errors	*/
		
	FIELD	rw, t1abs, t3abd, shf, 		/* Insn subfields					*/
			src, ccl, t1bbs, cin, 
			binv, iom, aid, decend, 
			erw, tdl, mrl, pac, 
			lsl, psc, flc, cir, 
			bcc, fls, baf, tbs, 
			bcf, ccm, mtsr, nma, 
			t1bbi, bcons, rmer;
} INSN;									/* Data to build an instruction		*/

typedef	struct {

	char	*srcn;					/* Current source file name				*/
	int		srcl;					/* Current source file lineno			*/
	Bool	error;					/* Error in entry						*/
	Bool	valid;					/* Memory word has assigned value		*/
	int		value;					/* Contents of the 16-bit word			*/
	int		line;					/* A line in the listfile				*/
	Bool	fset;					/* Function ID is set					*/
	int		func;					/* Entry's function identifier			*/
	char	*name;					/* Entry's name							*/
	Bool	fixed;					/* Entry start address is determined	*/
	int		start;					/* Start address						*/
	Bool	mena;					/* Match is enabled						*/
	Bool	magic;					/* Motorla magic						*/
	Bool	hsr0,					/* Condition: Host Service Request 0	*/
			hsr1,					/* Condition: Host Service Request 1	*/
			lsr,					/* Condition: Link Service Request		*/
			mtsr,					/* Condition: Match/Transition			*/
			pin,					/* Condition: Pin State					*/
			flag0;					/* Condition: Channel Flag 0			*/
	Bool	reg_p;					/* Preload P (True) or BIOD (False)		*/
	int		param;					/* Preload parameter number (PRM0..7)	*/
	
} ENTRY;							/* Function entry descriptor			*/

typedef	enum {
	
	TypeEmpty,							/* Word is empty					*/
	TypeEntry,							/* Word describes an entry slot		*/
	TypeInsn							/* Word describes an instruction	*/
	
} Type;									/* Memory word type					*/

typedef	struct {

	Bool	valid;						/* Memory word has assigned value	*/
	long	value;						/* Contents of the 32-bit word		*/
	int		line;						/* A line in the listfile			*/
	char	*srcn;						/* Current source file name			*/
	int		srcl;						/* Current source file lineno		*/
	INSN	insn;						/* Instruction for the word			*/
			
} WORD;									/* Memory word						*/

/****************************************************************************/
/*							Global data										*/
/****************************************************************************/

extern	WORD	tpuram[ MAXRAM ];		/* Image of the TPU RAM				*/
extern	ENTRY	entries[ MAXENT ];		/* Image of the TPU RAM				*/
extern	INSN	insn;					/* Global instruction descriptor	*/
extern	ENTRY	entry;					/* Global entry descriptor			*/
extern	int		pc;						/* Current uPC						*/
extern	int		maxmem;					/* Defines the max. TPU memory		*/
extern	int		def_function;			/* Default function					*/
extern	int		def_func_set;			/* Default function is valid		*/

extern	int		lineno;					/* Current list line number			*/
extern	int		src_lineno;				/* Current source line number		*/
extern	char	*sourcename;			/* Current source name				*/
extern	Bool	in_macro;				/* Currently processing a macro		*/
extern	FILE	*hexfile;				/* SREC file pointer				*/
extern	FILE	*listfile;				/* Listfile pointer					*/
extern	FILE	*incfile;				/* C include file pointer			*/
extern	Bool	lst_macro;				/* Flag: list macro expansions		*/
extern	char	elin[];					/* Error line						*/
extern	Bool	error;					/* Error occured					*/

extern	int		tab_size;				/* Size of the tab chars			*/
extern	int		max_line;				/* Max length of src lines listed	*/

extern	fpos_t	lstpos;					/* Position in the listfile			*/

# include "cinc.h"
# include "errors.h"
# include "insn.h"

void yyerror( char *s );
void InsnAssemble( void );
void DoPage( void );
void DoSetType( int n );
void DoOrg( int val );
void DoEntry( void );
void EntryFunc( int n );
void EntryName( IDENT *p );
void EntryStart( Bool constant, int value );
void EntryMatch( Bool n );
void EntryCond( Ematch which, Bool val );
void EntryRegp( Bool preg );
void EntryPar( int n );
void DoInclude( char *name );
void SetDefaultFunc( int n );
void SetBBSConstMaybe( int n );
void Token( const char *token, Bool macroname );
void AddIncludePath( char *s );
void InsnInitAll( void );
void EntryInit( void );
void ListfileInit( void );
void CreateListFile( char *name );
void FixupInsn( void );
void FixupEntry( void );
void SrecClose( void );
void ListCurrent( void );

#endif
